import aerosandbox as asb
import pytest


def test_conventional():
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.conventional import airplane
    analysis = asb.VortexLatticeMethod(
        airplane=airplane,
        op_point=asb.OperatingPoint(alpha=10),
    )
    return analysis.run()


def test_vanilla():
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.vanilla import airplane
    analysis = asb.VortexLatticeMethod(
        airplane=airplane,
        op_point=asb.OperatingPoint(alpha=10),
    )
    return analysis.run()


def test_flat_plate():
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.flat_plate import airplane
    analysis = asb.VortexLatticeMethod(
        airplane=airplane,
        op_point=asb.OperatingPoint(alpha=10),
    )
    return analysis.run()


def test_flat_plate_mirrored():
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.flat_plate_mirrored import airplane
    analysis = asb.VortexLatticeMethod(
        airplane=airplane,
        op_point=asb.OperatingPoint(alpha=10),
        spanwise_resolution=1,
        chordwise_resolution=3,
    )
    return analysis.run()


if __name__ == '__main__':
    # test_conventional()
    # test_vanilla()
    # test_flat_plate()['CL']
    test_flat_plate_mirrored()
    # pytest.main()



from aerosandbox import ExplicitAnalysis
from aerosandbox.geometry import *
from aerosandbox.performance import OperatingPoint
import aerosandbox.library.aerodynamics as aero
import aerosandbox.numpy as np
from aerosandbox.aerodynamics.aero_3D.aero_buildup_submodels import *


class AeroBuildup(ExplicitAnalysis):
    """
    A workbook-style aerodynamics buildup.
    """

    def __init__(self,
                 airplane: Airplane,
                 op_point: OperatingPoint,
                 ):
        ### Initialize
        self.airplane = airplane
        self.op_point = op_point

    def run(self):
        aero_components = []

        for wing in self.airplane.wings:
            aero = wing_aerodynamics(
                wing=wing.translate(-self.airplane.xyz_ref),
                op_point=self.op_point
            )
            aero_components.append(aero)

        for fuselage in self.airplane.fuselages:
            aero = fuselage_aerodynamics(
                fuselage=fuselage.translate(-self.airplane.xyz_ref),
                op_point=self.op_point
            )
            aero_components.append(aero)

        aero_total = {}

        for k in aero_components[0].keys():
            values = [
                component[k] for component in aero_components
            ]

            try:
                aero_total[k] = sum(values)
            except TypeError:
                aero_total[k] = [
                    sum([
                        value[i]
                        for value in values
                    ])
                    for i in range(3)
                ]

        ##### Add nondimensional forces, and nondimensional quantities.
        qS = self.op_point.dynamic_pressure() * self.airplane.s_ref

        aero_total["CL"] = aero_total["L"] / qS
        aero_total["CY"] = aero_total["Y"] / qS
        aero_total["CD"] = aero_total["D"] / qS
        aero_total["Cl"] = aero_total["l_b"] / qS / self.airplane.b_ref
        aero_total["Cm"] = aero_total["m_b"] / qS / self.airplane.c_ref
        aero_total["Cn"] = aero_total["n_b"] / qS / self.airplane.b_ref

        self.output = aero_total

        return aero_total


if __name__ == '__main__':
    from aerosandbox.aerodynamics.aero_3D.test_aero_3D.geometries.conventional import airplane

    aero = AeroBuildup(
        airplane=airplane,
        op_point=OperatingPoint(alpha=0, beta=1),
    ).run()

    from aerosandbox.tools.pretty_plots import plt, show_plot, contour, equal, set_ticks

    fig, ax = plt.subplots(2, 2)
    alpha = np.linspace(-10, 10, 1000)
    aero = AeroBuildup(
        airplane=airplane,
        op_point=OperatingPoint(
            velocity=100,
            alpha=alpha,
            beta=0
        ),
    ).run()

    plt.sca(ax[0, 0])
    plt.plot(alpha, aero["CL"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_L$")
    set_ticks(5, 1, 0.5, 0.1)

    plt.sca(ax[0, 1])
    plt.plot(alpha, aero["CD"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_D$")
    set_ticks(5, 1, 0.05, 0.01)
    plt.ylim(bottom=0)

    plt.sca(ax[1, 0])
    plt.plot(alpha, aero["Cm"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_m$")
    set_ticks(5, 1, 0.5, 0.1)

    plt.sca(ax[1, 1])
    plt.plot(alpha, aero["CL"] / aero["CD"])
    plt.xlabel(r"$\alpha$ [deg]")
    plt.ylabel(r"$C_L/C_D$")
    set_ticks(5, 1, 10, 2)

    show_plot(
        "`asb.AeroBuildup` Aircraft Aerodynamics"
    )

    fig, ax = plt.subplots(figsize=(7, 6))
    Beta, Alpha = np.meshgrid(np.linspace(-90, 90, 200), np.linspace(-90, 90, 200))
    aero = AeroBuildup(
        airplane=airplane,
        op_point=OperatingPoint(
            velocity=10,
            alpha=Alpha,
            beta=Beta
        ),
    ).run()
    contour(Beta, Alpha, aero["CL"], levels=30)
    equal()
    show_plot("AeroBuildup", r"$\beta$ [deg]", r"$\alpha$ [deg]")



import hashlib
import aerosandbox.numpy as np

def eng_string(
        x: float,
        unit: str = None,
        format='%.3g',
        si=True
) -> str:
    '''
    Taken from: https://stackoverflow.com/questions/17973278/python-decimal-engineering-notation-for-mili-10e-3-and-micro-10e-6/40691220

    Returns float/int value <x> formatted in a simplified engineering format -
    using an exponent that is a multiple of 3.

    Args:

        x: The value to be formatted. Float or int.

        unit: A unit of the quantity to be expressed, given as a string. Example: Newtons -> "N"

        format: A printf-style string used to format the value before the exponent.

        si: if true, use SI suffix for exponent. (k instead of e3, n instead of
            e-9, etc.)

    Examples:

    With format='%.2f':
        1.23e-08 -> 12.30e-9
             123 -> 123.00
          1230.0 -> 1.23e3
      -1230000.0 -> -1.23e6

    With si=True:
          1230.0 -> "1.23k"
      -1230000.0 -> "-1.23M"

    With unit="N" and si=True:
          1230.0 -> "1.23 kN"
      -1230000.0 -> "-1.23 MN"
    '''
    sign = ''
    if x < 0:
        x = -x
        sign = '-'
    elif x == 0:
        return format % 0
    elif np.isnan(x):
        return "NaN"
    exp = int(np.floor(np.log10(x)))
    exp3 = exp - (exp % 3)
    x3 = x / (10 ** exp3)

    if si and exp3 >= -24 and exp3 <= 24 and exp3 != 0:
        exp3_text = 'yzafpnμm kMGTPEZY'[(exp3 + 24) // 3]
    elif exp3 == 0:
        exp3_text = ''
    else:
        exp3_text = 'e%s' % exp3

    if unit is not None:
        if si:
            exp3_text = " " + exp3_text + unit
        else:
            exp3_text = exp3_text + " " + unit

    return ('%s' + format + '%s') % (sign, x3, exp3_text)


def latex_sci_notation_string(
        x: float,
        format='%.2e',
):
    """
    Converts a floating-point number to a LaTeX-style formatted string. Does not include the `$$` wrapping to put you in math mode.

    Does not use scientific notation if the base would be zero.

    Examples:

        latex_sci_notation_string(3000) -> '3 \\times 10^{3}'
    """
    float_str = format % x
    base, exponent = float_str.split("e")
    exponent = int(exponent)
    if exponent == 0:
        return base
    else:
        return r"{0} \times 10^{{{1}}}".format(base, exponent)


def hash_string(string: str) -> int:
    """
    Hashes a string into a quasi-random 32-bit integer! (Based on an MD5 checksum algorithm.)

    Usual warnings apply: it's MD5, don't use this for anything intended to be cryptographically secure.
    """
    md5 = hashlib.md5(string.encode('utf-8'))
    hash_hex = md5.hexdigest()
    hash_int = int(hash_hex, 16)
    hash_int64 = hash_int % (2 ** 32)
    return hash_int64



from aerosandbox.numpy.determine_type import *
import pytest
import numpy as np
import casadi as cas


def test_int():
    assert is_casadi_type(5, recursive=True) == False
    assert is_casadi_type(5, recursive=False) == False


def test_float():
    assert is_casadi_type(5., recursive=True) == False
    assert is_casadi_type(5., recursive=False) == False


def test_numpy():
    assert is_casadi_type(
        np.array([1, 2, 3]),
        recursive=True
    ) == False
    assert is_casadi_type(
        np.array([1, 2, 3]),
        recursive=False
    ) == False


def test_casadi():
    assert is_casadi_type(
        cas.GenMX_ones(5),
        recursive=False
    ) == True
    assert is_casadi_type(
        cas.GenMX_ones(5),
        recursive=True
    ) == True


def test_numpy_list():
    assert is_casadi_type(
        [np.array(5), np.array(7)],
        recursive=False
    ) == False
    assert is_casadi_type(
        [np.array(5), np.array(7)],
        recursive=True
    ) == False


def test_casadi_list():
    assert is_casadi_type(
        [cas.GenMX_ones(5), cas.GenMX_ones(5)],
        recursive=False
    ) == False
    assert is_casadi_type(
        [cas.GenMX_ones(5), cas.GenMX_ones(5)],
        recursive=True
    ) == True


def test_mixed_list():
    assert is_casadi_type(
        [np.array(5), cas.GenMX_ones(5)],
        recursive=False
    ) == False
    assert is_casadi_type(
        [np.array(5), cas.GenMX_ones(5)],
        recursive=True
    ) == True


def test_multi_level_contaminated_list():
    a = [[1 for _ in range(10)] for _ in range(10)]

    assert is_casadi_type(a, recursive=False) == False
    assert is_casadi_type(a, recursive=True) == False

    a[5][5] = cas.MX(1)

    assert is_casadi_type(a, recursive=False) == False
    assert is_casadi_type(a, recursive=True) == True

    a[5][5] = np.array(cas.DM(1), dtype="O")

    assert is_casadi_type(a, recursive=False) == False
    assert is_casadi_type(a, recursive=True) == False


if __name__ == '__main__':
    pytest.main([__file__])



import aerosandbox.numpy as np
from typing import Union


def calculate_induced_velocity_horseshoe(
        x_field: Union[float, np.ndarray],
        y_field: Union[float, np.ndarray],
        z_field: Union[float, np.ndarray],
        x_left: Union[float, np.ndarray],
        y_left: Union[float, np.ndarray],
        z_left: Union[float, np.ndarray],
        x_right: Union[float, np.ndarray],
        y_right: Union[float, np.ndarray],
        z_right: Union[float, np.ndarray],
        gamma: np.ndarray,
        trailing_vortex_direction: np.ndarray = None,
) -> [Union[float, np.ndarray], Union[float, np.ndarray], Union[float, np.ndarray]]:
    """
    Calculates the induced velocity at a point:
        [x_field, y_field, z_field]
    in a 3D potential-flow flowfield.

    In this flowfield, the following singularity elements are assumed:
        * A single horseshoe vortex consisting of a bound leg and two trailing legs



    Args:
        x_field:
        y_field:
        z_field:
        x_left:
        y_left:
        z_left:
        x_right:
        y_right:
        z_right:
        gamma:

    Returns:

    """
    if trailing_vortex_direction is None:
        trailing_vortex_direction = np.array([1, 0, 0])

    np.assert_equal_shape({
        "x_field": x_field,
        "y_field": y_field,
        "z_field": z_field,
    })
    np.assert_equal_shape({
        "x_left" : x_left,
        "y_left" : y_left,
        "z_left" : z_left,
        "x_right": x_right,
        "y_right": y_right,
        "z_right": z_right,
    })

    a_x = x_field - x_left
    a_y = y_field - y_left
    a_z = z_field - z_left

    b_x = x_field - x_right
    b_y = y_field - y_right
    b_z = z_field - z_right

    u_x = trailing_vortex_direction[0]
    u_y = trailing_vortex_direction[1]
    u_z = trailing_vortex_direction[2]

    # Handle the special case where the field point is on one of the legs
    def smoothed_inv(x):
        "Approximates 1/x with a function that sharply goes to 0 in the x -> 0 limit."
        return x / (x ** 2 + 1e-8)

    ### Do some useful arithmetic

    a_cross_b_x = a_y * b_z - a_z * b_y
    a_cross_b_y = a_z * b_x - a_x * b_z
    a_cross_b_z = a_x * b_y - a_y * b_x
    a_dot_b = a_x * b_x + a_y * b_y + a_z * b_z

    a_cross_u_x = a_y * u_z - a_z * u_y
    a_cross_u_y = a_z * u_x - a_x * u_z
    a_cross_u_z = a_x * u_y - a_y * u_x
    a_dot_u = a_x * u_x + a_y * u_y + a_z * u_z

    b_cross_u_x = b_y * u_z - b_z * u_y
    b_cross_u_y = b_z * u_x - b_x * u_z
    b_cross_u_z = b_x * u_y - b_y * u_x
    b_dot_u = b_x * u_x + b_y * u_y + b_z * u_z

    norm_a = (a_x ** 2 + a_y ** 2 + a_z ** 2) ** 0.5
    norm_b = (b_x ** 2 + b_y ** 2 + b_z ** 2) ** 0.5
    norm_a_inv = smoothed_inv(norm_a)
    norm_b_inv = smoothed_inv(norm_b)

    ### Calculate Vij

    term1 = (norm_a_inv + norm_b_inv) * smoothed_inv(norm_a * norm_b + a_dot_b)
    term2 = norm_a_inv * smoothed_inv(norm_a - a_dot_u)
    term3 = norm_b_inv * smoothed_inv(norm_b - b_dot_u)

    constant = gamma / (4 * np.pi)

    u = constant * (
            a_cross_b_x * term1 +
            a_cross_u_x * term2 -
            b_cross_u_x * term3
    )
    v = constant * (
            a_cross_b_y * term1 +
            a_cross_u_y * term2 -
            b_cross_u_y * term3
    )
    w = constant * (
            a_cross_b_z * term1 +
            a_cross_u_z * term2 -
            b_cross_u_z * term3
    )

    return u, v, w


if __name__ == '__main__':
    ##### Check single vortex
    u, v, w = calculate_induced_velocity_horseshoe(
        x_field=0,
        y_field=0,
        z_field=0,
        x_left=-1,
        y_left=-1,
        z_left=0,
        x_right=-1,
        y_right=1,
        z_right=0,
        gamma=1,
    )
    print(u, v, w)

    ##### Plot grid of single vortex
    args = (-2, 2, 30)
    x = np.linspace(*args)
    y = np.linspace(*args)
    z = np.linspace(*args)
    X, Y, Z = np.meshgrid(x, y, z)

    Xf = X.flatten()
    Yf = Y.flatten()
    Zf = Z.flatten()

    left = [0, -1, 0]
    right = [0, 1, 0]

    Uf, Vf, Wf = calculate_induced_velocity_horseshoe(
        x_field=Xf,
        y_field=Yf,
        z_field=Zf,
        x_left=left[0],
        y_left=left[1],
        z_left=left[2],
        x_right=right[0],
        y_right=right[1],
        z_right=right[2],
        gamma=1,
    )

    pos = np.stack((Xf, Yf, Zf)).T
    dir = np.stack((Uf, Vf, Wf)).T

    dir_norm = np.reshape(np.linalg.norm(dir, axis=1), (-1, 1))

    dir = dir / dir_norm * dir_norm ** 0.2

    import pyvista as pv

    pv.set_plot_theme('dark')
    plotter = pv.Plotter()
    plotter.add_arrows(
        cent=pos,
        direction=dir,
        mag=0.15
    )
    plotter.add_lines(
        lines=np.array([
            [Xf.max(), left[1], left[2]],
            left,
            right,
            [Xf.max(), right[1], right[2]]
        ])
    )
    plotter.show_grid()
    plotter.show()

    ##### Check multiple vortices
    args = (-2, 2, 30)
    x = np.linspace(*args)
    y = np.linspace(*args)
    z = np.linspace(*args)
    X, Y, Z = np.meshgrid(x, y, z)

    Xf = X.flatten()
    Yf = Y.flatten()
    Zf = Z.flatten()

    left = [0, -1, 0]
    center = [0, 0, 0]
    right = [0, 1, 0]

    lefts = np.array([left, center])
    rights = np.array([center, right])
    strengths = np.array([2, 1])


    def wide(array):
        return np.reshape(array, (1, -1))


    def tall(array):
        return np.reshape(array, (-1, 1))


    Uf_each, Vf_each, Wf_each = calculate_induced_velocity_horseshoe(
        x_field=wide(Xf),
        y_field=wide(Yf),
        z_field=wide(Zf),
        x_left=tall(lefts[:, 0]),
        y_left=tall(lefts[:, 1]),
        z_left=tall(lefts[:, 2]),
        x_right=tall(rights[:, 0]),
        y_right=tall(rights[:, 1]),
        z_right=tall(rights[:, 2]),
        gamma=tall(strengths),
    )

    Uf = np.sum(Uf_each, axis=0)
    Vf = np.sum(Vf_each, axis=0)
    Wf = np.sum(Wf_each, axis=0)

    pos = np.stack((Xf, Yf, Zf)).T
    dir = np.stack((Uf, Vf, Wf)).T

    dir_norm = np.reshape(np.linalg.norm(dir, axis=1), (-1, 1))

    dir = dir / dir_norm * dir_norm ** 0.2

    import pyvista as pv

    pv.set_plot_theme('dark')
    plotter = pv.Plotter()
    plotter.add_arrows(
        cent=pos,
        direction=dir,
        mag=0.15
    )
    plotter.add_lines(
        lines=np.array([
            [Xf.max(), left[1], left[2]],
            left,
            center,
            [Xf.max(), center[1], center[2]],
            center,
            right,
            [Xf.max(), right[1], right[2]]
        ])
    )
    plotter.show_grid()
    plotter.show()



### Import everything from NumPy

from numpy import *

### Overwrite some functions
from .array import *
from .arithmetic import *
from .calculus import *
from .conditionals import *
from .finite_difference_operators import *
from .integrate import *
from .interpolate import *
from .linalg_top_level import *
import aerosandbox.numpy.linalg as linalg
from .logicals import *
from .rotations import *
from .spacing import *
from .surrogate_model_tools import *
from .trig import *

### Force-overwrite built-in Python functions.

from numpy import round  # TODO check that min, max are properly imported



import aerosandbox as asb
import aerosandbox.numpy as np

sd7037 = asb.Airfoil("sd7037")

airplane = asb.Airplane(
    name="Vanilla",
    xyz_ref=[0.5, 0, 0],
    s_ref=9,
    c_ref=0.9,
    b_ref=10,
    wings=[
        asb.Wing(
            name="Wing",
            symmetric=True,
            xsecs=[
                asb.WingXSec(
                    xyz_le=[0, 0, 0],
                    chord=1,
                    twist=2,
                    airfoil=sd7037,
                ),
                asb.WingXSec(
                    xyz_le=[0.2, 5, 1],
                    chord=0.6,
                    twist=2,
                    airfoil=sd7037,
                )
            ]
        ),
        asb.Wing(
            name="H-stab",
            symmetric=True,
            xyz_le=[4, 0, 0],
            xsecs=[
                asb.WingXSec(
                    xyz_le=[0, 0, 0],
                    chord=0.7,
                ),
                asb.WingXSec(
                    xyz_le=[0.14, 1.25, 0],
                    chord=0.42
                ),
            ]
        ),
        asb.Wing(
            name="V-stab",
            xyz_le=[4, 0, 0],
            xsecs=[
                asb.WingXSec(
                    xyz_le=[0, 0, 0],
                    chord=0.7,
                ),
                asb.WingXSec(
                    xyz_le=[0.14, 0, 1],
                    chord=0.42
                )
            ]
        )
    ],
    fuselages=[
        asb.Fuselage(
            name="Fuselage",
            xyz_le=[0, 0, 0],
            xsecs=[
                asb.FuselageXSec(
                    xyz_c=[xi * 5 - 0.5, 0, 0],
                    radius=asb.Airfoil("naca0024").local_thickness(x_over_c=xi)
                )
                for xi in np.cosspace(0, 1, 30)
            ]
        )
    ]
)

if __name__ == '__main__':
    airplane.draw()



from aerosandbox.geometry.airfoil.airfoil_families import *
import pytest


def test_get_NACA_coordinates():
    coords = get_NACA_coordinates(
        name='naca4408',
        n_points_per_side=100
    )
    assert len(coords) == 199


def test_get_UIUC_coordinates():
    coords = get_UIUC_coordinates(
        name="dae11",
    )
    assert len(coords) != 0


if __name__ == '__main__':
    pytest.main()



"""
An illustration of the various fit types that can be used during function regression, and what they look like.
"""

from aerosandbox.modeling.fitting import FittedModel
import pytest
import aerosandbox.numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from dataset_temperature import time, measured_temperature


def test_fit_model_fit_type(plot=False):
    ### Fit a model
    def model(x, p):
        return p["2"] * x ** 2 + p["1"] * x + p["0"]  # Quadratic regression

    def fit_model_with_fit_type(fit_type):
        return FittedModel(
            model=model,
            x_data=time,
            y_data=measured_temperature,
            parameter_guesses={
                "2": 0,
                "1": 0,
                "0": 0,
            },
            fit_type=fit_type,
            residual_norm_type="L1"
        )

    best_fit_model = fit_model_with_fit_type("best")
    upper_bound_model = fit_model_with_fit_type("upper bound")
    lower_bound_model = fit_model_with_fit_type("lower bound")

    if plot:
        ### Plot fits with various different norms
        x = np.linspace(0, 100)

        sns.set(palette=sns.color_palette("husl", 3))
        fig, ax = plt.subplots(1, 1, figsize=(6.4, 4.8), dpi=200)
        plt.plot(time, measured_temperature, ".k", label="Data")
        plt.plot(x, best_fit_model(x), label=r"Best Fit")
        plt.plot(x, upper_bound_model(x), label=r"Upper-Bound Fit")
        plt.plot(x, lower_bound_model(x), label=r"Lower-Bound Fit")
        plt.xlabel(r"Time")
        plt.ylabel(r"Temperature")
        plt.title(r"Illustration of Fit Types for Robust Surrogate Modeling (Linear Model)")
        plt.tight_layout()
        plt.legend()
        plt.show()


if __name__ == '__main__':
    test_fit_model_fit_type(True)



import aerosandbox as asb
import aerosandbox.numpy as np

if __name__ == '__main__':
    af = asb.Airfoil("dae11", generate_polars=True)

    alpha = np.linspace(-40, 40, 300)
    re = np.geomspace(1e4, 1e12, 100)
    Alpha, Re = np.meshgrid(alpha, re)
    af.CL_function(alpha=0, Re=1e6)

    CL = af.CL_function(Alpha.flatten(), Re.flatten()).reshape(Alpha.shape)
    CD = af.CD_function(Alpha.flatten(), Re.flatten()).reshape(Alpha.shape)
    CM = af.CM_function(Alpha.flatten(), Re.flatten()).reshape(Alpha.shape)

    ##### Plot alpha-Re contours
    from aerosandbox.tools.pretty_plots import plt, show_plot, contour

    fig, ax = plt.subplots()
    contour(Alpha, Re, CL, levels=30)
    plt.scatter(af.xfoil_data["alpha"], af.xfoil_data["Re"], color="k", alpha=0.2)
    plt.yscale('log')
    show_plot()

    fig, ax = plt.subplots()
    contour(Alpha, Re, np.log10(CD), levels=30)
    plt.scatter(af.xfoil_data["alpha"], af.xfoil_data["Re"], color="k", alpha=0.2)
    plt.yscale('log')
    show_plot()

    ##### Plot Polars
    fig, ax = plt.subplots(2, 2, figsize=(8, 8))
    Re = 1e6
    alpha_lowres = np.linspace(-15, 15, 31)
    ma = alpha
    mCL = af.CL_function(alpha, Re)
    mCD = af.CD_function(alpha, Re)
    xf_run = asb.XFoil(af, Re=Re, max_iter=20).alpha(alpha_lowres)
    xa = xf_run["alpha"]
    xCL = xf_run["CL"]
    xCD = xf_run["CD"]

    plt.sca(ax[0, 0])
    plt.plot(ma, mCL)
    plt.plot(xa, xCL, ".")

    plt.sca(ax[0, 1])
    plt.plot(ma, mCD)
    plt.plot(xa, xCD, ".")
    plt.ylim(0, 0.05)

    plt.sca(ax[1, 0])
    plt.plot(mCD, mCL)
    plt.plot(xCD, xCL, ".")
    plt.xlim(0, 0.05)

    plt.sca(ax[1, 1])
    plt.plot(ma, mCL / mCD)
    plt.plot(xa, xCL / xCD, ".")

    show_plot()

    ##### Test optimization
    opti = asb.Opti()
    alpha = opti.variable(init_guess=0, lower_bound=-20, upper_bound=20)
    LD = af.CL_function(alpha, 1e6) / af.CD_function(alpha, 1e6)
    opti.minimize(
        -LD
    )
    sol = opti.solve()



from aerosandbox.modeling.interpolation import InterpolatedModel
import pytest
import aerosandbox.numpy as np


def underlying_function_2D(x1, x2):
    return x1 ** 2 + (x2 + 1) ** 2


def interpolated_model():
    np.random.seed(0)  # Set a seed for repeatability.

    ### Make some data
    x1 = np.linspace(0, 10, 11)
    x2 = np.linspace(0, 10, 21)

    X1, X2 = np.meshgrid(x1, x2, indexing="ij")

    return InterpolatedModel(
        x_data_coordinates={
            "x1": x1,
            "x2": x2,
        },
        y_data_structured=underlying_function_2D(X1, X2),
    )


def test_interpolated_model_at_scalar():
    model = interpolated_model()
    x_data = {
        "x1": 5.5,
        "x2": 5.5,
    }
    assert model(x_data) == pytest.approx(underlying_function_2D(*x_data.values()))


def test_interpolated_model_at_vector():
    model = interpolated_model()
    x_data = {
        "x1": np.array([1.5, 2.5]),
        "x2": np.array([2.5, 3.5]),
    }
    assert np.all(
        model(x_data) ==
        pytest.approx(underlying_function_2D(*x_data.values()))
    )


# def test_interpolated_model_plot(interpolated_model):
#     interpolated_model.plot()


if __name__ == '__main__':
    test_interpolated_model_at_scalar()
    test_interpolated_model_at_vector()
    # pytest.main()



import aerosandbox as asb
import aerosandbox.numpy as np
from aerosandbox.dynamics.dynamics import FreeBodyDynamics
import pytest

def test_quadcopter_navigation():

    opti = asb.Opti()

    N = 300
    time_final = 1
    time = np.linspace(0, time_final, N)

    left_thrust = opti.variable(init_guess=0.5, scale=1, n_vars=N, lower_bound=0, upper_bound=1)
    right_thrust = opti.variable(init_guess=0.5, scale=1, n_vars=N, lower_bound=0, upper_bound=1)

    mass = 0.1

    dyn = FreeBodyDynamics(
        opti_to_add_constraints_to=opti,
        time=time,
        xe=opti.variable(init_guess=np.linspace(0, 1, N)),
        ze=opti.variable(init_guess=np.linspace(0, -1, N)),
        u=opti.variable(init_guess=0, n_vars=N),
        w=opti.variable(init_guess=0, n_vars=N),
        theta=opti.variable(init_guess=np.linspace(np.pi / 2, np.pi / 2, N)),
        q=opti.variable(init_guess=0, n_vars=N),
        X=left_thrust + right_thrust,
        M=(right_thrust - left_thrust) * 0.1 / 2,
        mass=mass,
        Iyy=0.5 * mass * 0.1 ** 2,
        g=9.81,
    )

    opti.subject_to([  # Starting state
        dyn.xe[0] == 0,
        dyn.ze[0] == 0,
        dyn.u[0] == 0,
        dyn.w[0] == 0,
        dyn.theta[0] == np.radians(90),
        dyn.q[0] == 0,
        ])

    opti.subject_to([  # Final state
        dyn.xe[-1] == 1,
        dyn.ze[-1] == -1,
        dyn.u[-1] == 0,
        dyn.w[-1] == 0,
        dyn.theta[-1] == np.radians(90),
        dyn.q[-1] == 0,
        ])

    effort = np.sum( # The average "effort per second", where effort is integrated as follows:
        np.trapz(left_thrust ** 2 + right_thrust ** 2) * np.diff(time)
    ) / time_final

    opti.minimize(effort)

    sol = opti.solve()
    dyn.substitute_solution(sol)

    assert sol.value(effort) == pytest.approx(0.714563, rel=0.01)

    print(sol.value(effort))

def test_quadcopter_flip():

    opti = asb.Opti()

    N = 300
    time_final = opti.variable(init_guess=1, lower_bound=0)
    time = np.linspace(0, time_final, N)

    left_thrust = opti.variable(init_guess=0.7, scale=1, n_vars=N, lower_bound=0, upper_bound=1)
    right_thrust = opti.variable(init_guess=0.6, scale=1, n_vars=N, lower_bound=0, upper_bound=1)

    mass = 0.1

    dyn = FreeBodyDynamics(
        opti_to_add_constraints_to=opti,
        time=time,
        xe=opti.variable(init_guess=np.linspace(0, 1, N)),
        ze=opti.variable(init_guess=0, n_vars=N),
        u=opti.variable(init_guess=0, n_vars=N),
        w=opti.variable(init_guess=0, n_vars=N),
        theta=opti.variable(init_guess=np.linspace(np.pi / 2, np.pi / 2 - 2 * np.pi, N)),
        q=opti.variable(init_guess=0, n_vars=N),
        X=left_thrust + right_thrust,
        M=(right_thrust - left_thrust) * 0.1 / 2,
        mass=mass,
        Iyy=0.5 * mass * 0.1 ** 2,
        g=9.81,
    )

    opti.subject_to([  # Starting state
        dyn.xe[0] == 0,
        dyn.ze[0] == 0,
        dyn.u[0] == 0,
        dyn.w[0] == 0,
        dyn.theta[0] == np.radians(90),
        dyn.q[0] == 0,
        ])

    opti.subject_to([  # Final state
        dyn.xe[-1] == 1,
        dyn.ze[-1] == 0,
        dyn.u[-1] == 0,
        dyn.w[-1] == 0,
        dyn.theta[-1] == np.radians(90 - 360),
        dyn.q[-1] == 0,
        ])

    opti.minimize(time_final)

    sol = opti.solve(verbose=False)
    dyn.substitute_solution(sol)

    assert sol.value(time_final) == pytest.approx(0.824, abs=0.01)

if __name__ == '__main__':
    pytest.main()


import aerosandbox as asb
import aerosandbox.numpy as np
from aerosandbox.dynamics.utilities import inv_symmetric_3x3
import warnings


def equations_of_motion(
        xe=0,
        ye=0,
        ze=0,
        u=0,
        v=0,
        w=0,
        phi=0,
        theta=0,
        psi=0,
        p=0,
        q=0,
        r=0,
        X=0,
        Y=0,
        Z=0,
        L=0,
        M=0,
        N=0,
        mass=1,
        Ixx=1,
        Iyy=1,
        Izz=1,
        Ixy=0,
        Iyz=0,
        Ixz=0,
        g=0,
        hx=0,
        hy=0,
        hz=0,
):
    """
    Computes the state derivatives (i.e. equations of motion) for a body in 3D space.

    Based on Section 9.8.2 of Flight Vehicle Aerodynamics by Mark Drela.

    Args:
        xe: x-position, in Earth axes. [meters]
        ye: y-position, in Earth axes. [meters]
        ze: z-position, in Earth axes. [meters]
        u: x-velocity, in body axes. [m/s]
        v: y-velocity, in body axes. [m/s]
        w: z-velocity, in body axes. [m/s]
        phi: roll angle. Uses yaw-pitch-roll Euler angle convention. [rad]
        theta: pitch angle. Uses yaw-pitch-roll Euler angle convention. [rad]
        psi: yaw angle. Uses yaw-pitch-roll Euler angle convention. [rad]
        p: x-angular-velocity, in body axes. [rad/sec]
        q: y-angular-velocity, in body axes. [rad/sec]
        r: z-angular-velocity, in body axes. [rad/sec]
        X: x-direction force, in body axes. [N]
        Y: y-direction force, in body axes. [N]
        Z: z-direction force, in body axes. [N]
        L: Moment about the x axis, in body axes. Assumed these moments are applied about the center of mass. [Nm]
        M: Moment about the y axis, in body axes. Assumed these moments are applied about the center of mass. [Nm]
        N: Moment about the z axis, in body axes. Assumed these moments are applied about the center of mass. [Nm]
        mass: Mass of the body. [kg]
        Ixx: Respective component of the (symmetric) moment of inertia tensor.
        Iyy: Respective component of the (symmetric) moment of inertia tensor.
        Izz: Respective component of the (symmetric) moment of inertia tensor.
        Ixy: Respective component of the (symmetric) moment of inertia tensor.
        Ixz: Respective component of the (symmetric) moment of inertia tensor.
        Iyz: Respective component of the (symmetric) moment of inertia tensor.
        g: Magnitude of gravitational acceleration. Assumed to act in the positive-z-in-earth-axes ("downward") direction. [m/s^2]
        hx: x-component of onboard angular momentum (e.g. propellers), in body axes. [kg*m^2/sec]
        hy: y-component of onboard angular momentum (e.g. propellers), in body axes. [kg*m^2/sec]
        hz: z-component of onboard angular momentum (e.g. propellers), in body axes. [kg*m^2/sec]

    Returns:
        Time derivatives of each of the 12 state variables, given in a dictionary:
            {
                "xe"   : d_xe,
                "ye"   : d_ye,
                "ze"   : d_ze,
                "u"    : d_u,
                "v"    : d_v,
                "w"    : d_w,
                "phi"  : d_phi,
                "theta": d_theta,
                "psi"  : d_psi,
                "p"    : d_p,
                "q"    : d_q,
                "r"    : d_r,
            }
    """

    ### Trig Shorthands
    def sincos(x):
        try:
            x = np.mod(x, 2 * np.pi)
            one = np.ones_like(x)
            zero = np.zeros_like(x)

            if np.allclose(x, 0) or np.allclose(x, 2 * np.pi):
                sin = zero
                cos = one
            elif np.allclose(x, np.pi / 2):
                sin = one
                cos = zero
            elif np.allclose(x, np.pi):
                sin = zero
                cos = -one
            elif np.allclose(x, 3 * np.pi / 2):
                sin = -one
                cos = zero
            else:
                raise ValueError()
        except:
            sin = np.sin(x)
            cos = np.cos(x)
        return sin, cos

    # Do the trig
    sphi, cphi = sincos(phi)
    sthe, cthe = sincos(theta)
    spsi, cpsi = sincos(psi)

    ##### Equations of Motion

    ### Position derivatives
    d_xe = (
            (cthe * cpsi) * u +
            (sphi * sthe * cpsi - cphi * spsi) * v +
            (cphi * sthe * cpsi + sphi * spsi) * w
    )
    d_ye = (
            (cthe * sphi) * u +
            (sphi * sthe * spsi + cphi * cpsi) * v +
            (cphi * sthe * spsi - sphi * cpsi) * w
    )
    d_ze = (
            (-sthe) * u +
            (sphi * cthe) * v +
            (cphi * cthe) * w
    )
    ### Velocity derivatives
    d_u = (
            (X / mass - g * sthe) -
            q * w +
            r * v
    )
    d_v = (
            (Y / mass + g * sphi * cthe) -
            r * u +
            p * w
    )
    d_w = (
            (Z / mass + g * cphi * cthe) -
            p * v +
            q * u
    )
    ### Angle derivatives
    if np.any(cthe != 0):
        d_phi = (
                p +
                q * sphi * sthe / cthe +
                r * cphi * sthe / cthe
        )
    else:
        d_phi = 0

    d_theta = (
            q * cphi -
            r * sphi
    )

    if np.any(cthe != 0):
        d_psi = (
            q * sphi / cthe +
            r * cphi / cthe
        )
    else:
        d_psi = 0

    ### Angular velocity derivatives
    RHS_L = (
            L -
            (Izz - Iyy) * q * r -
            Iyz * (q ** 2 - r ** 2) -
            Ixz * p * q +
            Ixy * p * r -
            hz * q +
            hy * r
    )
    RHS_M = (
            M -
            (Ixx - Izz) * r * p -
            Ixz * (r ** 2 - p ** 2) -
            Ixy * q * r +
            Iyz * q * p -
            hx * r +
            hz * p
    )
    RHS_N = (
            N -
            (Iyy - Ixx) * p * q -
            Ixy * (p ** 2 - q ** 2) -
            Iyz * r * p +
            Ixz * r * q -
            hy * p +
            hx * q
    )
    i11, i22, i33, i12, i23, i13 = inv_symmetric_3x3(Ixx, Iyy, Izz, Ixy, Iyz, Ixz)

    d_p = i11 * RHS_L + i12 * RHS_M + i13 * RHS_N
    d_q = i12 * RHS_L + i22 * RHS_M + i23 * RHS_N
    d_r = i13 * RHS_L + i23 * RHS_M + i33 * RHS_N

    return {
        "xe"   : d_xe,
        "ye"   : d_ye,
        "ze"   : d_ze,
        "u"    : d_u,
        "v"    : d_v,
        "w"    : d_w,
        "phi"  : d_phi,
        "theta": d_theta,
        "psi"  : d_psi,
        "p"    : d_p,
        "q"    : d_q,
        "r"    : d_r,
    }



def inv_symmetric_3x3(
        m11,
        m22,
        m33,
        m12,
        m23,
        m13,
):
    """
    Explicitly computes the inverse of a symmetric 3x3 matrix.

    Input matrix (note symmetry):
    
    [m11, m12, m13]
    [m12, m22, m23]
    [m13, m23, m33]
    
    Output matrix (note symmetry):
    
    [a11, a12, a13]
    [a12, a22, a23]
    [a13, a23, a33]

    From https://math.stackexchange.com/questions/233378/inverse-of-a-3-x-3-covariance-matrix-or-any-positive-definite-pd-matrix
    """
    inv_det = 1 / (
            m11 * (m33 * m22 - m23 ** 2) -
            m12 * (m33 * m12 - m23 * m13) +
            m13 * (m23 * m12 - m22 * m13)
    )
    a11 = m33 * m22 - m23 ** 2
    a12 = m13 * m23 - m33 * m12
    a13 = m12 * m23 - m13 * m22

    a22 = m33 * m11 - m13 ** 2
    a23 = m12 * m13 - m11 * m23

    a33 = m11 * m22 - m12 ** 2

    a11 = a11 * inv_det
    a12 = a12 * inv_det
    a13 = a13 * inv_det
    a22 = a22 * inv_det
    a23 = a23 * inv_det
    a33 = a33 * inv_det

    return a11, a22, a33, a12, a23, a13



"""
A set of tools used for making prettier Matplotlib plots.
"""

import matplotlib.pyplot as plt
import seaborn as sns
import matplotlib as mpl
from typing import Union, Dict, List, Callable
from matplotlib import ticker
import aerosandbox.numpy as np
from aerosandbox.tools.string_formatting import eng_string

# plt.ion()

### Define color palettes
palettes = {
    "categorical": [
        "#4285F4",  # From Google logo, blue
        "#EA4335",  # From Google logo, red
        "#34A853",  # From Google logo, green
        "#ECB22E",  # From Slack logo, gold
        "#9467BD",  # Rest are from Matplotlib "tab10"
        "#8C564B",
        "#E377C2",
        "#7F7F7F",
    ],
}

sns.set_theme(
    palette=palettes["categorical"],
)

mpl.rcParams["figure.dpi"] = 200
mpl.rcParams["axes.formatter.useoffset"] = False
mpl.rcParams["contour.negative_linestyle"] = 'solid'

def set_ticks(
        x_major: Union[float, int] = None,
        x_minor: Union[float, int] = None,
        y_major: Union[float, int] = None,
        y_minor: Union[float, int] = None
):
    ax = plt.gca()
    if x_major is not None:
        ax.xaxis.set_major_locator(ticker.MultipleLocator(base=x_major))
    if x_minor is not None:
        ax.xaxis.set_minor_locator(ticker.MultipleLocator(base=x_minor))
    if y_major is not None:
        ax.yaxis.set_major_locator(ticker.MultipleLocator(base=y_major))
    if y_minor is not None:
        ax.yaxis.set_minor_locator(ticker.MultipleLocator(base=y_minor))

def equal():
    plt.gca().set_aspect("equal", adjustable='box')

def adjust_lightness(color, amount=1.0):
    """
    Converts a color to HLS space, then mulitplies the lightness by `amount`, then converts back to RGB.

    Args:
        color: A color, in any format that matplotlib understands.
        amount: The amount to multiply the lightness by. Valid range is 0 to infinity.

    Returns: A color, as an RGB tuple.

    """
    import matplotlib.colors as mc
    import colorsys
    try:
        c = mc.cnames[color]
    except:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))
    return colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), c[2])


def show_plot(
        title: str = None,
        xlabel: str = None,
        ylabel: str = None,
        tight_layout: bool = True,
        legend: bool = None,
        legend_frame: bool = True,
        show: bool = True,
        pretty_grids: bool = True,
):
    """
    Finalize and show a plot.
    Args:
        title:
        xlabel:
        ylabel:
        tight_layout:
        legend:
        show:
        pretty_grids:

    Returns:

    """
    fig = plt.gcf()
    axes = fig.get_axes()

    if pretty_grids:
        for ax in axes:
            ax.grid(True, 'major', axis='both', linewidth=1.6)
            ax.grid(True, 'minor', axis='both', linewidth=0.7)

    ### Determine if a legend should be shown
    if legend is None:
        lines = plt.gca().lines
        if len(lines) <= 1:
            legend = False
        else:
            legend = False
            for line in lines:
                if line.get_label()[0] != "_":
                    legend = True
                    break

    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)
    if title is not None:
        if len(axes) <= 1:
            plt.title(title)
        else:
            plt.suptitle(title)
    if tight_layout:
        plt.tight_layout()
    if legend:
        plt.legend(frameon=legend_frame)
    if show:
        plt.show()


def contour(
        X,
        Y,
        Z,
        levels: Union[int, List, np.ndarray] = None,
        colorbar=True,
        linelabels=True,
        cmap=mpl.cm.get_cmap('viridis'),
        alpha: float = 0.7,
        extend="both",
        linecolor="k",
        linewidths=0.5,
        extendrect=True,
        linelabels_format: Union[str, Callable[[float], str]] = eng_string,
        linelabels_fontsize=8,
        colorbar_label=None,
        contour_kwargs: Dict = None,
        contourf_kwargs: Dict = None,
        colorbar_kwargs: Dict = None,
        linelabels_kwargs: Dict = None,
        **kwargs,
):
    """
    An analogue for plt.contour and plt.tricontour and friends that produces a much prettier default graph.

    Can take inputs with either contour or tricontour syntax.

    See syntax here:
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.contour.html
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.contourf.html
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tricontour.html
        https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.tricontourf.html

    Args:
        X: See contour docs.
        Y: See contour docs.
        Z: See contour docs.
        levels: See contour docs.
        colorbar: Should we draw a colorbar?
        linelabels: Should we add line labels?
        cmap: What colormap should we use?
        alpha: What transparency should all plot elements be?
        extend: See contour docs.
        linecolor: What color should the line labels be?
        linewidths: See contour docs.
        extendrect: See colorbar docs.
        linelabels_format: See ax.clabel docs.
        linelabels_fontsize: See ax.clabel docs.
        contour_kwargs: Additional keyword arguments for contour.
        contourf_kwargs: Additional keyword arguments for contourf.
        colorbar_kwargs: Additional keyword arguments for colorbar.
        linelabels_kwargs: Additional keyword arguments for the line labels (ax.clabel).
        **kwargs: Additional keywords, assumed to be for both contour and contourf.

    Returns: A tuple of (contour, contourf, colorbar) objects.

    """

    if contour_kwargs is None:
        contour_kwargs = {}
    if contourf_kwargs is None:
        contourf_kwargs = {}
    if colorbar_kwargs is None:
        colorbar_kwargs = {}
    if linelabels_kwargs is None:
        linelabels_kwargs = {}

    args = [
        X,
        Y,
        Z,
    ]

    shared_kwargs = kwargs
    if levels is not None:
        shared_kwargs["levels"] = levels
    shared_kwargs["alpha"] = alpha
    shared_kwargs["extend"] = extend

    if colorbar_label is not None:
        colorbar_kwargs["label"] = colorbar_label

    contour_kwargs = {
        "colors"    : linecolor,
        "linewidths": linewidths,
        **shared_kwargs,
        **contour_kwargs
    }
    contourf_kwargs = {
        "cmap": cmap,
        **shared_kwargs,
        **contourf_kwargs
    }

    colorbar_kwargs = {
        "extendrect": extendrect,
        **colorbar_kwargs
    }

    linelabels_kwargs = {
        "inline"  : 1,
        "fontsize": linelabels_fontsize,
        "fmt"     : linelabels_format,
        **linelabels_kwargs
    }

    try:
        cont = plt.contour(*args, **contour_kwargs)
        contf = plt.contourf(*args, **contourf_kwargs)
    except TypeError as e:
        try:
            cont = plt.tricontour(*args, **contour_kwargs)
            contf = plt.tricontourf(*args, **contourf_kwargs)
        except TypeError:
            raise e

    if colorbar:
        cbar = plt.colorbar(**colorbar_kwargs)
    else:
        cbar = None

    if linelabels:
        plt.gca().clabel(cont, **linelabels_kwargs)

    return cont, contf, cbar


### A few functions to take a WAG at the *technological* feasibility of
#   a nuclear-RTG-powered airplane. Not at all endorsing this as a good idea,
#   just a fascinating one that's surprisingly technologically feasible with Polonium-210.

# Aircraft uses:
# There's speculation that the Russians might already be flying this?
# Source (Overview): https://theconversation.com/nuclear-powered-missile-accident-in-russia-what-really-happened-121966
# Source (Russian press release): http://rosatom.ru/journalist/news/zayavlenie-departamenta-kommunikatsiy-goskorporatsii-rosatom/
# Source (American perspective): https://foreignpolicy.com/2019/08/12/russia-mysterious-explosion-arctic-putin-chernobyl/

# How it's made:
# Po-210 is created by neutron bombardment of (stable) bismuth-209 in a nuclear reactor.
# This forms radioactive bismuth-210 (half-life: 5 days), which decays to Po-210 via beta decay.
# Source: https://personal.ems.psu.edu/~radovic/Polonium_specs.pdf

# Cost:
# Po-210 appears to be affordable, at least in milligram amounts (used in manufacturing
# processes for things as simple as static elimination).
# Source: https://personal.ems.psu.edu/~radovic/Polonium_specs.pdf

# Safety:
# Alpha emitter, so safe as long as it's not ingested or inhaled. If ingested, it has a LD50 of 0.7 micrograms.
# If inhaled, it's about 20% of that.
# Source: https://personal.ems.psu.edu/~radovic/Polonium_specs.pdf

# Supply:
# Only 100 grams of Po-210 is produced each year, almost all in Russia.
# Source: https://www.nrc.gov/reading-rm/doc-collections/fact-sheets/polonium.html
# This seems to be an insurmountable issue.

alpha_particle_mass_amu = 4.001506179127  # mass of an alpha particle
po_210_mass_amu = 209.9828736  # mass of Polonium-210
pb_206_mass_amu = 205.9744818  # mass of Lead-206
amu = 1.6605390666050e-27  # kg, 1 atomic mass unit
c = 299792458  # m/s, speed of light


def po210_specific_power(
        days_after_formation=0,
):
    half_life = 138.376  # days
    # Source: https://en.wikipedia.org/wiki/Polonium-210
    pure_specific_energy = (
            (po_210_mass_amu - alpha_particle_mass_amu - pb_206_mass_amu) * c ** 2
            / po_210_mass_amu
    )  # J/kg
    # TODO finish



from aerosandbox import *
import pytest


def test_init():  # TODO actually test this
    xsec = WingXSec(
        xyz_le=np.array([0, 0, 0]),
        chord=1.,
        twist=0,
        airfoil=Airfoil("naca0012"),
        control_surface_is_symmetric=True,
        control_surface_hinge_point=0.75,
        control_surface_deflection=0.,
    )


if __name__ == '__main__':
    pytest.main()



import warnings
import copy
import aerosandbox.numpy as np

class Material:
    
    # A short list of properties, more can be added
    
    # Physical Properties
    density                    = None  # kg/m^3
    
    # Mechanical Properties
    tensile_strength_ultimate  = None  # Pa
    tensile_strength_yield     = None  # Pa
    elongation_at_break        = None  # Percentage
    modulus_of_elasticity      = None
    bulk_modulus               = None
    poissons_ratio              = None
    shear_modulus              = None
    
    # Electrical Properties
    electrical_conductivity     = None
    
    # Thermal Properties
    cte_linear                 = None
    specific_heat_capacity     = None
    thermal_conductivity       = None
    
    # Extras
    isotropic                  = None

    # add redirects for easy access? Maybe a bad idea
    _aliases = {
        'density': [
            'rho',
        ],
        'tensile_strength_yield': [
            'yield_strength',
            'yield_stress',
        ],
        'tensile_strength_Ultimate':[
            'ultimate_strength',
            'tensile_strength',  # Not sure I like this, but typical?
        ],
        'elongation_at_break': [
            'elongation',
        ],
        'poissons_ratio': [
            'poisson_ratio'
        ],
        'modulus_of_elasticity':[
            'elastic_modulus',
            'E',
        ],
        'electrical_conductivity': [
            'elastic_modulus'
            'conductivity',
        ],
        'shear_modulus': [
            'modulus_of_rigidity',
            'G',
        ],
        'cte_linear': [
            'coefficient_of_thermal_expansion',
        ],
        'specific_heat_capacity': [
            'specific_heat',
        ]
        }
    
    def _set_aliases(self):
        for key in self._aliases.keys():
            #redirect aliases to original
            
            def make_getter(key):
                def getter(self):
                    # print('getting ' + key)
                    return getattr(self, key)
                return getter
            
            def make_setter(key):
                def setter(self, value):
                    # print('setting ' + key)
                    setattr(self, key, value)
                return setter
            
            for alias in self._aliases[key]:
                prop = property(fget=make_getter(key),
                                fset=make_setter(key))
                
                # Add it to the list of properties
                setattr(self.__class__, alias, prop)      
            
    
    def __init__(self, **kwargs):
        
        self._set_aliases()
        
        # Just stuff  everything into vars for now
        for key in kwargs.keys():
            setattr(self, key, kwargs[key])
      

# %% Web Scraping
      
# A class to import directly from MatMatch? Not 100% reliable, but the lazy way
#class MatMatch_Material(Material):
#    
#    def __init__(self, 
#                 url: str,
#                 force_value: str = None  # min/max
#                 ):
#        self.force_value = force_value
#        
#        warnings.warn('This is a web scraping interface.'+\
#                      '\nAs it depends on an external resource,'+\
#                      ' functionality is not guaranteed.'+\
#                      '\nUse at your own risk.')
#        
#        import requests
#        import json
#        try:
#            from bs4 import BeautifulSoup
#        except ImportError:
#            raise ImportError('Requires BeautifulSoup 4:'+
#                              '/n    pip install beautifulsoup4'+
#                              '/n(not installed by default)')
#        
#        # Get the webpage
#        r = requests.get(url)
#        soup = BeautifulSoup(r.content, 'html.parser')
#        
#        # Hackish interception of the data loading script
#        content = soup.find('script', {'id': '__NEXT_DATA__'}).contents
#        
#        # Get the json data out
#        jsoncontent = json.loads(content[0])
#        
#        # Get the material data
#        sections = jsoncontent['props']['initialProps']['pageProps']['data']['sections']
#        
#        # Get the properties section
#        sections_dict = {x['sectionId']:x for x in sections}
#        
#        # Consolidate all  the sections
#        properties = {}
#        for section in sections_dict['Properties']['sections']:
#            for entry in section['properties']:
#                name = entry['metadata']['name']
#                
#                # make variable format
#                name = name.lower().replace(' ', '_')
#                name = name.replace("'", "")
#                
#                for datapoint in entry['dataPoints']:
#                    if datapoint['metadata']['unitsSystem'] == 'metric':
#                        value_metric = datapoint['value']
#                        units = datapoint['metadata']['unit']
#                    
#                properties.update({name:{'value': value_metric,
#                                         'units': units}})
#                
#        # TODO: Convert to SI (might require pint?)
#        properties = self._convert_to_SI(properties)
#        self.p1 = properties
#        
#        # TODO: Figure out min max
#        properties = self._check_min_max(properties)
#        self.p2 = properties
#        
#        kwargs = {key: properties[key]
#                  for key in properties}
#        
#        super().__init__(**kwargs)
#        
#    @staticmethod
#    def _convert_to_SI(properties):
#        
#        # This can very likely be done much better
#        # Factors are the mul needed to get to SI, same for offsets
#        known_units = {
#            '°C':       {'offset': 272.15},
#            'Ω·m':      {},
#            'W/(m·K)':  {},
#            'J/(kg·K)': {},
#            '1/K':      {},
#            'MPa':      {'factor': 10E6},
#            'GPa':      {'factor': 10E9},
#            'kPa':      {'factor': 10E3},
#            'Pa':       {},
#            '%':        {'factor': 1/100},
#            '[-]':      {},
#            'g/cm³':    {'factor': 10E3},
#            'kg/m³':    {}
#            }
#        
#        props_out = copy.deepcopy(properties)
#        for key in properties.keys():
#            values = properties[key]['value']
#            units = properties[key]['units']
#            
#            if units not in known_units.keys():
#                warnings.warn('Unknown units: ' + units + '\n'\
#                              'No known conversion, these may not be SI')
#                val = ''
#                while val.lower() != 'y' and val.lower() != 'n':
#                    val = input('Include without converting? (y/n) ')
#                    
#                if val.lower() == 'n':
#                    print('Skipping ' + key)
#                    props_out.pop(key)
#                    continue
#                else:
#                    props_out[key] = {'min': values['min'],
#                                       'max': values['max']}
#                    continue
#            
#            mini = values['min'] * known_units[units].get('factor', 1) + \
#                known_units[units].get('offset', 0)
#            maxi = values['max'] * known_units[units].get('factor', 1) + \
#                known_units[units].get('offset', 0)
#                
#            # Let's drop temp for now
#            props_out[key] = {'min': mini, 'max': maxi}
#            
#        return props_out
#    
#    def _check_min_max(self, properties):
#        
#        for key in properties.keys():
#            val = properties[key]
#            
#            if type(self.force_value) == str:
#                force_value = self.force_value
#            elif type(self.force_value) == dict:
#                force_value = self.force_value[key]
#            
#            if force_value == 'max':
#                properties[key] = val['max']
#                continue
#            elif force_value == 'min':
#                properties[key] = val['min']
#                continue
#            elif val['max'] == val['min']:
#                properties[key] = val['max']
#                continue
#            else:
#                warnings.warn('Min/max mismatch for: ' + key + '\n'\
#                              'Please specify value, or force_values in init.')
#                str_in = ''
#                
#                while str_in.lower() != 'min' \
#                    and str_in.lower() != 'max' \
#                    and not np.char.isnumeric(str_in):
#                    str_in = input('Specify value (min/max/float) ')
#                    
#                if np.char.isnumeric(str_in):
#                    properties[key] = float(str_in)
#                    continue
#                if str_in.lower() == 'min':
#                    properties[key] = val['min']
#                    continue
#                if str_in.lower() == 'max':
#                    properties[key] = val['max']
#                    continue
#                else:
#                    input('waddup')
#            
#        return properties
#            
#            
#class AISI_1006_Steel_Cold_Drawn(MatMatch_Material):
#    
#    def __init__(self):
#        url = 'https://matmatch.com/materials/mitf545-aisi-1006-cold-drawn'
#        
#        super().__init__(url, force_value='min')
        


import aerosandbox.numpy as np


def induced_drag(
        lift,
        span,
        dynamic_pressure,
        oswalds_efficiency=1,
):
    """
    Computes the induced drag associated with a lifting planar wing.

    Args:
        lift: Lift force [Newtons]
        span: Wing span [meters]
        dynamic_pressure: Dynamic pressure [Pascals]
        oswalds_efficiency: Oswald's efficiency factor [-]

    Returns: Induced drag force [Newtons]

    """
    return lift ** 2 / (
            dynamic_pressure * np.pi * span ** 2 * oswalds_efficiency
    )


def oswalds_efficiency(
        taper_ratio: float,
        aspect_ratio: float,
        sweep: float = 0.,
        fuselage_diameter_to_span_ratio: float = 0.,
) -> float:
    """
    Computes the Oswald's efficiency factor for a planar, tapered, swept wing.

    Based on "Estimating the Oswald Factor from Basic Aircraft Geometrical Parameters"
    by M. Nita, D. Scholz; Hamburg Univ. of Applied Sciences, 2012.

    Implementation of Section 5 from the above paper.

    Only valid for backwards-swept wings; i.e. 0 <= sweep < 90.

    Args:
        taper_ratio: Taper ratio of the wing (tip_chord / root_chord) [-]
        aspect_ratio: Aspect ratio of the wing (b^2 / S) [-]
        sweep: Wing quarter-chord sweep angle [deg]

    Returns: Oswald's efficiency factor [-]

    """
    sweep = np.clip(sweep, 0, 90)  # TODO input proper analytic continuation

    def f(l):  # f(lambda), given as Eq. 36 in the Nita and Scholz paper (see parent docstring).
        return (
                0.0524 * l ** 4
                - 0.15 * l ** 3
                + 0.1659 * l ** 2
                - 0.0706 * l
                + 0.0119
        )

    delta_lambda = -0.357 + 0.45 * np.exp(-0.0375 * sweep)
    # Eq. 37 in Nita & Scholz.
    # Note: there is a typo in the cited paper; the negative in the exponent was omitted.
    # A bit of thinking about this reveals that this omission must be erroneous.

    e_theo = 1 / (
            1 + f(taper_ratio - delta_lambda) * aspect_ratio
    )

    fuselage_wake_contraction_correction_factor = 1 - 2 * (fuselage_diameter_to_span_ratio) ** 2

    e = e_theo * fuselage_wake_contraction_correction_factor

    return e


def optimal_taper_ratio(
        sweep=0.,
) -> float:
    """
    Computes the optimal (minimum-induced-drag) taper ratio for a given quarter-chord sweep angle.

    Based on "Estimating the Oswald Factor from Basic Aircraft Geometrical Parameters"
    by M. Nita, D. Scholz; Hamburg Univ. of Applied Sciences, 2012.

    Only valid for backwards-swept wings; i.e. 0 <= sweep < 90.

    Args:
        sweep: Wing quarter-chord sweep angle [deg]

    Returns: Optimal taper ratio

    """
    sweep = np.clip(sweep, 0, 90)  # TODO input proper analytic continuation
    return 0.45 * np.exp(-0.0375 * sweep)


def CL_over_Cl(
        aspect_ratio: float,
        mach: float = 0.,
        sweep: float = 0.
) -> float:
    """
    Returns the ratio of 3D lift coefficient (with compressibility) to 2D lift coefficient (incompressible).
    :param aspect_ratio: Aspect ratio
    :param mach: Mach number
    :param sweep: Sweep angle [deg]
    :return:
    """
    beta_squared = np.sqrt(
        (1 - mach) ** 2 + 0.02
    )
    # return aspect_ratio / (aspect_ratio + 2) # Equivalent to equation in Drela's FVA in incompressible, 2*pi*alpha limit.
    # return aspect_ratio / (2 + np.sqrt(4 + aspect_ratio ** 2))  # more theoretically sound at low aspect_ratio
    eta = 0.95
    return aspect_ratio / (
            2 + np.sqrt(
        4 + (
                aspect_ratio ** 2 * beta_squared / eta ** 2
        ) * (
                1 + np.tand(sweep) ** 2 / beta_squared
        )
    )
    )  # From Raymer, Sect. 12.4.1; citing DATCOM


def induced_drag_ratio_from_ground_effect(
        h_over_b  # type: float
):
    """
    Gives the ratio of actual induced drag to free-flight induced drag experienced by a wing in ground effect.
    Artificially smoothed below around h/b == 0.05 to retain differentiability and practicality.
    Source: W. F. Phillips, D. F. Hunsaker, "Lifting-Line Predictions for Induced Drag and Lift in Ground Effect".
        Using Equation 5 from the paper, which is modified from a model from Torenbeek:
            Torenbeek, E. "Ground Effects", 1982.
    :param h_over_b: (Height above ground) divided by (wingspan).
    :return: Ratio of induced drag in ground effect to induced drag out of ground effect [unitless]
    """
    h_over_b = np.softmax(
        h_over_b,
        0,
        hardness=1 / 0.03
    )
    return 1 - np.exp(
        -4.01 * (2 * h_over_b) ** 0.717
    )



import aerosandbox as asb
import aerosandbox.numpy as np

if __name__ == '__main__':
    af1 = asb.Airfoil("naca0012")
    af1.generate_polars(cache_filename="airfoil_data_cache.json")
    af2 = asb.Airfoil("naca0012")
    af2.generate_polars(cache_filename="airfoil_data_cache.json")



from aerosandbox.geometry import *
from aerosandbox.performance import OperatingPoint
import aerosandbox.numpy as np
import aerosandbox.library.aerodynamics as aerolib


def wing_aerodynamics(
        wing: Wing,
        op_point: OperatingPoint,
):
    """
    Estimates the aerodynamic forces, moments, and derivatives on a wing in isolation.

    Moments are given with the reference at Wing [0, 0, 0].

    Args:

        wing: A Wing object that you wish to analyze.

        op_point: The OperatingPoint that you wish to analyze the fuselage at.

    TODO account for wing airfoil pitching moment

    Returns:

    """
    ##### Compute general wing properties and things to be used in sectional analysis.
    sweep = wing.mean_sweep_angle()
    AR = wing.aspect_ratio()
    mach = op_point.mach()
    q = op_point.dynamic_pressure()
    CL_over_Cl = aerolib.CL_over_Cl(
        aspect_ratio=AR,
        mach=mach,
        sweep=sweep
    )
    oswalds_efficiency = aerolib.oswalds_efficiency(
        taper_ratio=wing.taper_ratio(),
        aspect_ratio=AR,
        sweep=sweep,
        fuselage_diameter_to_span_ratio=0  # an assumption
    )
    areas = wing.area(_sectional=True)
    aerodynamic_centers = wing.aerodynamic_center(_sectional=True)

    F_g = [0, 0, 0]
    M_g = [0, 0, 0]

    ##### Iterate through the wing sections.
    for sect_id in range(len(wing.xsecs) - 1):

        ##### Identify the wing cross sections adjacent to this wing section.
        xsec_a = wing.xsecs[sect_id]
        xsec_b = wing.xsecs[sect_id + 1]

        ##### When linearly interpolating, weight things by the relative chord.
        a_weight = xsec_a.chord / (xsec_a.chord + xsec_b.chord)
        b_weight = xsec_b.chord / (xsec_a.chord + xsec_b.chord)

        ##### Compute the local frame of this section, and put the z (normal) component into wind axes.
        xg_local, yg_local, zg_local = wing._compute_frame_of_section(sect_id)
        sect_aerodynamic_center = aerodynamic_centers[sect_id]

        sect_z_w = op_point.convert_axes(
            x_from=zg_local[0], y_from=zg_local[1], z_from=zg_local[2],
            from_axes="geometry",
            to_axes="wind",
        )

        ##### Compute the generalized angle of attack, so the geometric alpha that the wing section "sees".
        velocity_vector_b_from_freestream = op_point.convert_axes(
            x_from=-op_point.velocity, y_from=0, z_from=0,
            from_axes="wind",
            to_axes="body"
        )
        velocity_vector_b_from_rotation = np.cross(
            op_point.convert_axes(
                sect_aerodynamic_center[0], sect_aerodynamic_center[1], sect_aerodynamic_center[2],
                from_axes="geometry",
                to_axes="body"
            ),
            [op_point.p, op_point.q, op_point.r],
            manual=True
        )
        velocity_vector_b = [
            velocity_vector_b_from_freestream[i] + velocity_vector_b_from_rotation[i]
            for i in range(3)
        ]
        velocity_mag_b = np.sqrt(sum([comp ** 2 for comp in velocity_vector_b]))
        velocity_dir_b = [
            velocity_vector_b[i] / velocity_mag_b
            for i in range(3)
        ]
        sect_z_b = op_point.convert_axes(
            x_from=zg_local[0], y_from=zg_local[1], z_from=zg_local[2],
            from_axes="geometry",
            to_axes="body",
        )
        vel_dot_normal = np.dot(velocity_dir_b, sect_z_b, manual=True)

        sect_alpha_generalized = 90 - np.arccosd(vel_dot_normal)

        ##### Compute sectional lift at cross sections using lookup functions. Merge them linearly to get section CL.
        xsec_a_Cl_incompressible = xsec_a.airfoil.CL_function(
            alpha=sect_alpha_generalized,
            Re=op_point.reynolds(xsec_a.chord),
            mach=0,  # Note: this is correct, mach correction happens in 2D -> 3D step
            deflection=xsec_a.control_surface_deflection  # TODO treat symmetry
        )
        xsec_b_Cl_incompressible = xsec_b.airfoil.CL_function(
            alpha=sect_alpha_generalized,
            Re=op_point.reynolds(xsec_b.chord),
            mach=0,  # Note: this is correct, mach correction happens in 2D -> 3D step
            deflection=xsec_a.control_surface_deflection
        )
        sect_CL = (
                          xsec_a_Cl_incompressible * a_weight +
                          xsec_b_Cl_incompressible * b_weight
                  ) * CL_over_Cl

        ##### Compute sectional drag at cross sections using lookup functions. Merge them linearly to get section CD.
        xsec_a_Cd_profile = xsec_a.airfoil.CD_function(
            alpha=sect_alpha_generalized,
            Re=op_point.reynolds(xsec_a.chord),
            mach=mach,
            deflection=xsec_a.control_surface_deflection
        )
        xsec_b_Cd_profile = xsec_b.airfoil.CD_function(
            alpha=sect_alpha_generalized,
            Re=op_point.reynolds(xsec_b.chord),
            mach=mach,
            deflection=xsec_a.control_surface_deflection
        )
        sect_CDp = (
                xsec_a_Cd_profile * a_weight +
                xsec_b_Cd_profile * b_weight
        )

        ##### Compute induced drag from local CL and full-wing properties (AR, e)
        sect_CDi = (
                sect_CL ** 2 / (np.pi * AR * oswalds_efficiency)
        )

        ##### Total the drag.
        sect_CD = sect_CDp + sect_CDi

        ##### Go to dimensional quantities using the area.
        area = areas[sect_id]
        sect_L = q * area * sect_CL
        sect_D = q * area * sect_CD

        ##### Compute the direction of the lift by projecting the section's normal vector into the plane orthogonal to the freestream.
        sect_L_direction_w = (
            np.zeros_like(sect_z_w[0]),
            sect_z_w[1] / np.sqrt(sect_z_w[1] ** 2 + sect_z_w[2] ** 2),
            sect_z_w[2] / np.sqrt(sect_z_w[1] ** 2 + sect_z_w[2] ** 2)
        )
        sect_L_direction_g = op_point.convert_axes(
            *sect_L_direction_w, from_axes="wind", to_axes="geometry"
        )

        ##### Compute the direction of the drag by aligning the drag vector with the freestream vector.
        sect_D_direction_w = (-1, 0, 0)
        sect_D_direction_g = op_point.convert_axes(
            *sect_D_direction_w, from_axes="wind", to_axes="geometry"
        )

        ##### Compute the force vector in geometry axes.
        sect_F_g = [
            sect_L * sect_L_direction_g[i] + sect_D * sect_D_direction_g[i]
            for i in range(3)
        ]

        ##### Compute the moment vector in geometry axes.
        sect_M_g = np.cross(
            sect_aerodynamic_center,
            sect_F_g,
            manual=True
        )

        ##### Add section forces and moments to overall forces and moments
        F_g = [
            F_g[i] + sect_F_g[i]
            for i in range(3)
        ]
        M_g = [
            M_g[i] + sect_M_g[i]
            for i in range(3)
        ]

        ##### Treat symmetry
        if wing.symmetric:
            ##### Compute the local frame of this section, and put the z (normal) component into wind axes.

            sym_sect_aerodynamic_center = aerodynamic_centers[sect_id]
            sym_sect_aerodynamic_center[1] *= -1

            sym_sect_z_w = op_point.convert_axes(
                x_from=zg_local[0], y_from=-zg_local[1], z_from=zg_local[2],
                from_axes="geometry",
                to_axes="wind",
            )

            ##### Compute the generalized angle of attack, so the geometric alpha that the wing section "sees".
            sym_velocity_vector_b_from_freestream = op_point.convert_axes(
                x_from=-op_point.velocity, y_from=0, z_from=0,
                from_axes="wind",
                to_axes="body"
            )
            sym_velocity_vector_b_from_rotation = np.cross(
                op_point.convert_axes(
                    sym_sect_aerodynamic_center[0], sym_sect_aerodynamic_center[1], sym_sect_aerodynamic_center[2],
                    from_axes="geometry",
                    to_axes="body"
                ),
                [op_point.p, op_point.q, op_point.r],
                manual=True
            )
            sym_velocity_vector_b = [
                sym_velocity_vector_b_from_freestream[i] + sym_velocity_vector_b_from_rotation[i]
                for i in range(3)
            ]
            sym_velocity_mag_b = np.sqrt(sum([comp ** 2 for comp in sym_velocity_vector_b]))
            sym_velocity_dir_b = [
                sym_velocity_vector_b[i] / sym_velocity_mag_b
                for i in range(3)
            ]
            sym_sect_z_b = op_point.convert_axes(
                x_from=zg_local[0], y_from=-zg_local[1], z_from=zg_local[2],
                from_axes="geometry",
                to_axes="body",
            )
            sym_vel_dot_normal = np.dot(sym_velocity_dir_b, sym_sect_z_b, manual=True)

            sym_sect_alpha_generalized = 90 - np.arccosd(sym_vel_dot_normal)

            ##### Compute sectional lift at cross sections using lookup functions. Merge them linearly to get section CL.
            sym_xsec_a_Cl_incompressible = xsec_a.airfoil.CL_function(
                alpha=sym_sect_alpha_generalized,
                Re=op_point.reynolds(xsec_a.chord),
                mach=0,  # Note: this is correct, mach correction happens in 2D -> 3D step
                deflection=xsec_a.control_surface_deflection * (1 if xsec_a.control_surface_is_symmetric else -1)
            )
            sym_xsec_b_Cl_incompressible = xsec_b.airfoil.CL_function(
                alpha=sym_sect_alpha_generalized,
                Re=op_point.reynolds(xsec_b.chord),
                mach=0,  # Note: this is correct, mach correction happens in 2D -> 3D step
                deflection=xsec_a.control_surface_deflection * (1 if xsec_a.control_surface_is_symmetric else -1)
            )
            sym_sect_CL = (
                                  sym_xsec_a_Cl_incompressible * a_weight +
                                  sym_xsec_b_Cl_incompressible * b_weight
                          ) * CL_over_Cl

            ##### Compute sectional drag at cross sections using lookup functions. Merge them linearly to get section CD.
            sym_xsec_a_Cd_profile = xsec_a.airfoil.CD_function(
                alpha=sym_sect_alpha_generalized,
                Re=op_point.reynolds(xsec_a.chord),
                mach=mach,
                deflection=xsec_a.control_surface_deflection * (1 if xsec_a.control_surface_is_symmetric else -1)
            )
            sym_xsec_b_Cd_profile = xsec_b.airfoil.CD_function(
                alpha=sym_sect_alpha_generalized,
                Re=op_point.reynolds(xsec_b.chord),
                mach=mach,
                deflection=xsec_a.control_surface_deflection * (1 if xsec_a.control_surface_is_symmetric else -1)
            )
            sym_sect_CDp = (
                    sym_xsec_a_Cd_profile * a_weight +
                    sym_xsec_b_Cd_profile * b_weight
            )

            ##### Compute induced drag from local CL and full-wing properties (AR, e)
            sym_sect_CDi = (
                    sym_sect_CL ** 2 / (np.pi * AR * oswalds_efficiency)
            )

            ##### Total the drag.
            sym_sect_CD = sym_sect_CDp + sym_sect_CDi

            ##### Go to dimensional quantities using the area.
            area = areas[sect_id]
            sym_sect_L = q * area * sym_sect_CL
            sym_sect_D = q * area * sym_sect_CD

            ##### Compute the direction of the lift by projecting the section's normal vector into the plane orthogonal to the freestream.
            sym_sect_L_direction_w = (
                np.zeros_like(sym_sect_z_w[0]),
                sym_sect_z_w[1] / np.sqrt(sym_sect_z_w[1] ** 2 + sym_sect_z_w[2] ** 2),
                sym_sect_z_w[2] / np.sqrt(sym_sect_z_w[1] ** 2 + sym_sect_z_w[2] ** 2)
            )
            sym_sect_L_direction_g = op_point.convert_axes(
                *sym_sect_L_direction_w, from_axes="wind", to_axes="geometry"
            )

            ##### Compute the direction of the drag by aligning the drag vector with the freestream vector.
            sym_sect_D_direction_w = (-1, 0, 0)
            sym_sect_D_direction_g = op_point.convert_axes(
                *sym_sect_D_direction_w, from_axes="wind", to_axes="geometry"
            )

            ##### Compute the force vector in geometry axes.
            sym_sect_F_g = [
                sym_sect_L * sym_sect_L_direction_g[i] + sym_sect_D * sym_sect_D_direction_g[i]
                for i in range(3)
            ]

            ##### Compute the moment vector in geometry axes.
            sym_sect_M_g = np.cross(
                sym_sect_aerodynamic_center,
                sym_sect_F_g,
                manual=True
            )

            ##### Add section forces and moments to overall forces and moments
            F_g = [
                F_g[i] + sym_sect_F_g[i]
                for i in range(3)
            ]
            M_g = [
                M_g[i] + sym_sect_M_g[i]
                for i in range(3)
            ]

    ##### Convert F_g and M_g to body and wind axes for reporting.
    F_b = op_point.convert_axes(*F_g, from_axes="geometry", to_axes="body")
    F_w = op_point.convert_axes(*F_b, from_axes="body", to_axes="wind")
    M_b = op_point.convert_axes(*M_g, from_axes="geometry", to_axes="body")
    M_w = op_point.convert_axes(*M_b, from_axes="body", to_axes="wind")

    return {
        "F_g": F_g,
        "F_b": F_b,
        "F_w": F_w,
        "M_g": M_g,
        "M_b": M_b,
        "M_w": M_w,
        "L"  : -F_w[2],
        "Y"  : F_w[1],
        "D"  : -F_w[0],
        "l_b": M_b[0],
        "m_b": M_b[1],
        "n_b": M_b[2]
    }


if __name__ == '__main__':
    wing = Wing(
        xyz_le=[1, 0, 0],
        xsecs=[
            WingXSec(
                xyz_le=[0, 0, 0],
                chord=1,
                airfoil=Airfoil("naca0012"),
                twist=0,
            ),
            WingXSec(
                xyz_le=[0.5, 1, 0],
                chord=0.5,
                airfoil=Airfoil("naca0012"),
                twist=0,
            ),
            # WingXSec(
            #     xyz_le=[0.7, 1, 0.3],
            #     chord=0.3,
            #     airfoil=Airfoil("naca0012"),
            #     twist=0,
            # )
        ],
        symmetric=True
    )
    aero = wing_aerodynamics(
        wing=wing,
        op_point=OperatingPoint(
            velocity=10,
            alpha=10,
            beta=0,
            p=1
        )
    )
    print(aero)



from aerosandbox.dynamics.dynamics import FreeBodyDynamics
import aerosandbox as asb
import aerosandbox.numpy as np

import pytest


def test_block_move_fixed_time():
    opti = asb.Opti()

    n_timesteps = 300

    time = np.linspace(0, 1, n_timesteps)

    dyn = FreeBodyDynamics(
        opti_to_add_constraints_to=opti,
        time=time,
        xe=opti.variable(init_guess=np.linspace(0, 1, n_timesteps)),
        u=opti.variable(init_guess=1, n_vars=n_timesteps),
        X=opti.variable(init_guess=np.linspace(1, -1, n_timesteps)),
        mass=1,
    )

    opti.subject_to([
        dyn.xe[0] == 0,
        dyn.xe[-1] == 1,
        dyn.u[0] == 0,
        dyn.u[-1] == 0,
    ])

    # effort = np.sum(
    #     np.trapz(dyn.X ** 2) * np.diff(time)
    # )

    effort = np.sum(  # More sophisticated integral-of-squares integration (closed form correct)
        np.diff(time) / 3 *
        (dyn.X[:-1] ** 2 + dyn.X[:-1] * dyn.X[1:] + dyn.X[1:] ** 2)
    )

    opti.minimize(effort)

    sol = opti.solve()

    dyn.substitute_solution(sol)

    assert dyn.xe[0] == pytest.approx(0)
    assert dyn.xe[-1] == pytest.approx(1)
    assert dyn.u[0] == pytest.approx(0)
    assert dyn.u[-1] == pytest.approx(0)
    assert np.max(dyn.u) == pytest.approx(1.5, abs=0.01)
    assert dyn.X[0] == pytest.approx(6, abs=0.05)
    assert dyn.X[-1] == pytest.approx(-6, abs=0.05)


def test_block_move_minimum_time():
    opti = asb.Opti()

    n_timesteps = 300

    time = np.linspace(
        0,
        opti.variable(init_guess=1, lower_bound=0),
        n_timesteps,
    )

    dyn = FreeBodyDynamics(
        opti_to_add_constraints_to=opti,
        time=time,
        xe=opti.variable(init_guess=np.linspace(0, 1, n_timesteps)),
        u=opti.variable(init_guess=1, n_vars=n_timesteps),
        X=opti.variable(init_guess=np.linspace(1, -1, n_timesteps), lower_bound=-1, upper_bound=1),
        mass=1,
    )

    opti.subject_to([
        dyn.xe[0] == 0,
        dyn.xe[-1] == 1,
        dyn.u[0] == 0,
        dyn.u[-1] == 0,
    ])

    opti.minimize(
        time[-1]
    )

    sol = opti.solve()

    dyn.substitute_solution(sol)

    assert dyn.xe[0] == pytest.approx(0)
    assert dyn.xe[-1] == pytest.approx(1)
    assert dyn.u[0] == pytest.approx(0)
    assert dyn.u[-1] == pytest.approx(0)
    assert np.max(dyn.u) == pytest.approx(1, abs=0.01)
    assert dyn.X[0] == pytest.approx(1, abs=0.05)
    assert dyn.X[-1] == pytest.approx(-1, abs=0.05)
    assert np.mean(np.abs(dyn.X)) == pytest.approx(1, abs=0.01)


def test_rocket_primitive():
    ### Parameters
    N = 100  # Number of discretization points
    time_final = 100  # seconds
    time = np.linspace(0, time_final, N)

    ### Constants
    mass_initial = 500e3  # Initial mass, 500 metric tons
    ze_final = -100e3  # Final position, 100 km
    g = 9.81  # Gravity, m/s^2
    alpha = 1 / (300 * g)  # kg/(N*s), Inverse of specific impulse, basically - don't worry about this

    ### Environment
    opti = asb.Opti()

    ### Variables
    dyn = FreeBodyDynamics(
        opti_to_add_constraints_to=opti,
        time=time,
        ze=opti.variable(init_guess=np.linspace(0, ze_final, N)),  # Earth-axis z, or "negative altitude"
        u=opti.variable(init_guess=-ze_final / time_final, n_vars=N),  # Velocity
        theta=np.pi / 2,  # Point the rocket up
        mass=opti.variable(init_guess=mass_initial, n_vars=N),  # Mass
        X=opti.variable(init_guess=g * mass_initial, n_vars=N),  # Thrust force (control vector)
        g=9.81,
    )

    ### Dynamics (implemented manually for now, we'll show you more sophisticated ways to do this in the Trajectory
    # Optimization part of the tutorial later on)
    opti.subject_to([  # Forward Euler, implemented manually for now
        np.diff(dyn.mass) == np.trapz(-alpha * dyn.X) * np.diff(time)
    ])

    ### Boundary conditions
    opti.subject_to([
        dyn.ze[0] == 0,
        dyn.ze[-1] == ze_final,
        dyn.u[0] == 0,
        dyn.mass[0] == mass_initial,
    ])

    ### Path constraints
    opti.subject_to([
        dyn.mass >= 0,
        dyn.X >= 0
    ])

    ### Objective
    opti.minimize(-dyn.mass[-1])  # Maximize the final mass == minimize fuel expenditure

    ### Solve
    sol = opti.solve()

    assert sol.value(dyn.mass[-1]) == pytest.approx(290049.81034472014, rel=0.05)
    assert sol.value(dyn.u).max() == pytest.approx(1448, rel=0.05)

def test_rocket_with_data_structures():

    ### Environment
    opti = asb.Opti()

    ### Time discretization
    N = 100  # Number of discretization points
    time_final = 100  # seconds
    time = np.linspace(0, time_final, N)

    ### Constants
    mass_initial = 500e3  # Initial mass, 500 metric tons
    ze_final = -100e3  # Final altitude, 100 km
    g = 9.81  # Gravity, m/s^2
    alpha = 1 / (300 * g)  # kg/(N*s), Inverse of specific impulse, basically - don't worry about this

    dyn = FreeBodyDynamics(
        opti_to_add_constraints_to=opti,
        time=time,
        ze=opti.variable(init_guess=np.linspace(0, ze_final, N)),  # Altitude (negative due to Earth-axes convention)
        u=opti.variable(init_guess=-ze_final / time_final, n_vars=N),  # Velocity
        theta=np.pi / 2,
        X=opti.variable(init_guess=g * mass_initial, n_vars=N),  # Mass
        mass=opti.variable(init_guess=mass_initial, n_vars=N),  # Control vector
        g=9.81,
    )

    opti.constrain_derivative(
        derivative=-alpha * dyn.X,
        variable=dyn.mass,
        with_respect_to=dyn.time,
    )

    ### Boundary conditions
    opti.subject_to([
        dyn.ze[0] == 0,
        dyn.u[0] == 0,
        dyn.mass[0] == mass_initial,
        dyn.ze[-1] == ze_final,
        ])

    ### Path constraints
    opti.subject_to([
        dyn.mass >= 0,
        dyn.X >= 0
    ])

    ### Objective
    opti.minimize(-dyn.mass[-1])  # Maximize the final mass == minimize fuel expenditure

    ### Solve
    sol = opti.solve(verbose=False)
    print(f"Solved in {sol.stats()['iter_count']} iterations.")

    assert sol.value(dyn.mass[-1]) == pytest.approx(290049.81034472014, rel=0.05)
    assert sol.value(dyn.u).max() == pytest.approx(1448, rel=0.05)

if __name__ == '__main__':
    pytest.main()



import aerosandbox.numpy as np
import casadi as cas
import pytest


def test_sum():
    a = np.arange(101)

    assert np.sum(a) == 5050  # Gauss would be proud.


def test_sum2():
    # Check it returns the same results with casadi and numpy
    a = np.array([[1, 2, 3], [1, 2, 3]])
    b = cas.SX(a)

    assert np.all(np.sum(a) == cas.DM(np.sum(b)))
    assert np.all(np.sum(a, axis=1) == cas.DM(np.sum(b, axis=1)))


def test_mean():
    a = np.linspace(0, 10, 50)

    assert np.mean(a) == pytest.approx(5)


def test_cumsum():
    n = np.arange(6).reshape((3, 2))
    c = cas.DM(n)

    assert np.all(
        np.cumsum(n) == np.array([0, 1, 3, 6, 10, 15])
    )
    # assert np.all( # TODO add casadi testing here
    #     np.cumsum(c) == np.array([0, 1, 3, 6, 10, 15])
    # )


if __name__ == '__main__':
    pytest.main()



from aerosandbox.common import *
from aerosandbox import Opti
import aerosandbox.numpy as np
from aerosandbox.dynamics.equations_of_motion import equations_of_motion
from aerosandbox import OperatingPoint, Atmosphere
import warnings


class FreeBodyDynamics(AeroSandboxObject):
    def __init__(self,
                 time: np.ndarray,
                 xe: np.ndarray = None,
                 ye: np.ndarray = None,
                 ze: np.ndarray = None,
                 u: np.ndarray = None,
                 v: np.ndarray = None,
                 w: np.ndarray = None,
                 phi: np.ndarray = None,
                 theta: np.ndarray = None,
                 psi: np.ndarray = None,
                 p: np.ndarray = None,
                 q: np.ndarray = None,
                 r: np.ndarray = None,
                 X=0,
                 Y=0,
                 Z=0,
                 L=0,
                 M=0,
                 N=0,
                 mass=1,
                 Ixx=1,
                 Iyy=1,
                 Izz=1,
                 Ixy=0,
                 Iyz=0,
                 Ixz=0,
                 g=0,
                 hx=0,
                 hy=0,
                 hz=0,
                 opti_to_add_constraints_to: Opti= None,
                 ):

        self.time = time
        self.xe = 0 if xe is None else xe
        self.ye = 0 if ye is None else ye
        self.ze = 0 if ze is None else ze
        self.u = 0 if u is None else u
        self.v = 0 if v is None else v
        self.w = 0 if w is None else w
        self.phi = 0 if phi is None else phi
        self.theta = 0 if theta is None else theta
        self.psi = 0 if psi is None else psi
        self.p = 0 if p is None else p
        self.q = 0 if q is None else q
        self.r = 0 if r is None else r
        self.X = X
        self.Y = Y
        self.Z = Z
        self.L = L
        self.M = M
        self.N = N
        self.mass = mass
        self.Ixx = Ixx
        self.Iyy = Iyy
        self.Izz = Izz
        self.Ixy = Ixy
        self.Iyz = Iyz
        self.Ixz = Ixz
        self.g = g
        self.hx = hx
        self.hy = hy
        self.hz = hz

        if opti_to_add_constraints_to is not None:
            state = self.state
            state_derivatives = self.state_derivatives()
            for k in state.keys():  # TODO default to second-order integration for position, angles
                if locals()[k] is None:  # Don't constrain states that haven't been defined by the user.
                    continue
                try:
                    opti_to_add_constraints_to.constrain_derivative(
                        derivative=state_derivatives[k],
                        variable=state[k],
                        with_respect_to=self.time,
                    )
                except Exception as e:
                    raise ValueError(f"Error while constraining state variable `{k}`: \n{e}")

    def __repr__(self):
        repr = []
        repr.append("Dynamics instance:")

        def trim(string, width=40):
            string = string.strip()
            if len(string) > width:
                return string[:width - 3] + "..."
            else:
                return string

        def makeline(k, v):
            name = trim(str(k), width=8).rjust(8)
            item = trim(str(v), width=40).ljust(40)

            try:
                value = str(self.opti.value(v))
            except:
                value = None

            if str(value).strip() == str(item).strip():
                value = None

            if isinstance(v, float) or isinstance(v, int) or isinstance(v, np.ndarray):
                value = None

            if value is not None:
                value = trim(value, width=40).ljust(40)
                return f"\t\t{name}: {item} ({value})"
            else:
                return f"\t\t{name}: {item}"

        repr.append("\tState variables:")
        for k, v in self.state.items():
            repr.append(makeline(k, v))

        repr.append("\tControl/other variables:")
        for k, v in self.control_variables.items():
            repr.append(makeline(k, v))

        return "\n".join(repr)

    # TODO add __getitem__ for dynamic state at instant in time

    @property
    def state(self):
        return {
            "xe"   : self.xe,
            "ye"   : self.ye,
            "ze"   : self.ze,
            "u"    : self.u,
            "v"    : self.v,
            "w"    : self.w,
            "phi"  : self.phi,
            "theta": self.theta,
            "psi"  : self.psi,
            "p"    : self.p,
            "q"    : self.q,
            "r"    : self.r,
        }

    def state_derivatives(self):
        return equations_of_motion(
            xe=self.xe,
            ye=self.ye,
            ze=self.ze,
            u=self.u,
            v=self.v,
            w=self.w,
            phi=self.phi,
            theta=self.theta,
            psi=self.psi,
            p=self.p,
            q=self.q,
            r=self.r,
            X=self.X,
            Y=self.Y,
            Z=self.Z,
            L=self.L,
            M=self.M,
            N=self.N,
            mass=self.mass,
            Ixx=self.Ixx,
            Iyy=self.Iyy,
            Izz=self.Izz,
            Ixy=self.Ixy,
            Iyz=self.Iyz,
            Ixz=self.Ixz,
            g=self.g,
            hx=self.hx,
            hy=self.hy,
            hz=self.hz,
        )

    @property
    def control_variables(self):
        return {
            "X"       : self.X,
            "Y"       : self.Y,
            "Z"       : self.Z,
            "L"       : self.L,
            "M"       : self.M,
            "N"       : self.N,
            "mass"    : self.mass,
            "Ixx"     : self.Ixx,
            "Iyy"     : self.Iyy,
            "Izz"     : self.Izz,
            "Ixy"     : self.Ixy,
            "Iyz"     : self.Iyz,
            "Ixz"     : self.Ixz,
            "g"       : self.g,
            "hx"      : self.hx,
            "hy"      : self.hy,
            "hz"      : self.hz,
            "alpha"   : self.alpha,
            "beta"    : self.beta,
            "speed"   : self.speed,
            "altitude": self.altitude
        }

    @property
    def alpha(self):
        """The angle of attack, in degrees."""
        return np.arctan2d(
            self.w,
            self.u
        )

    @property
    def beta(self):
        """The sideslip angle, in degrees."""
        return np.arctan2d(
            self.v,
            (self.u ** 2 + self.w ** 2) ** 0.5
        )

    @property
    def speed(self):
        """The speed of the object, expressed as a scalar."""
        return (
                       self.u ** 2 +
                       self.v ** 2 +
                       self.w ** 2
               ) ** 0.5

    @property
    def altitude(self):
        return -self.ze

    @property
    def op_point(self):
        return OperatingPoint(
            atmosphere=Atmosphere(altitude=-self.ze),
            velocity=self.speed,
            alpha=self.alpha,
            beta=self.beta,
            p=self.p,
            q=self.q,
            r=self.r,
        )

    def convert_axes(self,
                     x_from, y_from, z_from,
                     from_axes: str,
                     to_axes: str,
                     ):
        """
        Converts a vector [x_from, y_from, z_from], as given in the `from_axes` frame, to an equivalent vector [x_to,
        y_to, z_to], as given in the `to_axes` frame.

        Identical to OperatingPoint.convert_axes(), but adds in "earth" as a valid axis frame. For more documentation,
        see the docstring of OperatingPoint.convert_axes().

        Both `from_axes` and `to_axes` should be a string, one of:
                * "geometry"
                * "body"
                * "wind"
                * "stability"
                * "earth"

        Args:
                x_from: x-component of the vector, in `from_axes` frame.
                y_from: y-component of the vector, in `from_axes` frame.
                z_from: z-component of the vector, in `from_axes` frame.
                from_axes: The axes to convert from.
                to_axes: The axes to convert to.

        Returns: The x-, y-, and z-components of the vector, in `to_axes` frame. Given as a tuple.

        """
        try:
            x_b, y_b, z_b = self.op_point.convert_axes(
                x_from, y_from, z_from,
                from_axes=from_axes, to_axes="body"
            )
        except ValueError:
            if from_axes == "earth":
                do_earth_thing() # TODO DO
            else:
                raise ValueError("Bad value of `from_axes`!")

        try:
            x_to, y_to, z_to = self.op_point.convert_axes(
                x_b, y_b, z_b,
                from_axes="body", to_axes=to_axes
            )
        except ValueError:
            if to_axes == "earth":
                do_earth_thing() # TODO DO
            else:
                raise ValueError("Bad value of `to_axes`!")

        return x_to, y_to, z_to


if __name__ == '__main__':
    import aerosandbox as asb

    opti = asb.Opti()

    n_timesteps = 300

    time = np.linspace(0, 1, n_timesteps)

    dyn = FreeBodyDynamics(
        opti=opti,
        time=time,
        xe=opti.variable(init_guess=np.linspace(0, 1, n_timesteps)),
        u=opti.variable(init_guess=1, n_vars=n_timesteps),
        X=opti.variable(init_guess=np.linspace(1, -1, n_timesteps)),
    )

    opti.subject_to([
        dyn.xe[0] == 0,
        dyn.xe[-1] == 1,
        dyn.u[0] == 0,
        dyn.u[-1] == 0,
    ])

    opti.minimize(
        np.sum(np.trapz(dyn.X ** 2) * np.diff(time))
    )

    sol = opti.solve()

    dyn.substitute_solution(sol)



import aerosandbox as asb
import aerosandbox.numpy as np
import pytest


def test_warm_start():
    ### Set up an optimization problem

    opti = asb.Opti()

    x = opti.variable(init_guess=10)

    opti.minimize(
        x ** 4
    )

    ### Now, solve the optimization problem and print out how many iterations were needed to solve it.

    sol = opti.solve(verbose=False)
    n_iterations_1 = sol.stats()['iter_count']
    print(f"First run: Solved in {n_iterations_1} iterations.")

    ### Then, set the initial guess of the Opti problem to the solution that was found.
    opti.set_initial_from_sol(sol)

    ### Then, re-solve the problem and print how many iterations were now needed.
    sol = opti.solve(verbose=False)
    n_iterations_2 = sol.stats()['iter_count']
    print(f"Second run: Solved in {n_iterations_2} iterations.")

    ### Assert that fewer iterations were needed after a warm-start.
    assert n_iterations_2 < n_iterations_1


if __name__ == '__main__':
    pytest.main()
